API_BASE_URL = "https://test.eodatahub.org.uk"
S3_BASE_URL_TEMPLATE = (
    "https://{workspace_name}.workspaces.test.eodhp.eco-ke-staging.com/files/"
    "eodhp-test-workspaces1"
)
PAGINATION_LIMIT = 10
